from django.shortcuts import render, get_object_or_404
from django.views.generic import ListView, DetailView, FormView
from django.views.generic.edit import CreateView, UpdateView, DeleteView
from django.contrib.auth.mixins import LoginRequiredMixin, UserPassesTestMixin
from django.http import QueryDict
from django import forms
from django.contrib.postgres.search import SearchQuery, SearchHeadline
from django.core.exceptions import PermissionDenied

from django.http import HttpResponseRedirect
from django.urls import reverse
import urllib

from .forms import EventSubmissionForm, EventForm, BatchImportationForm

from .models import Event, Category, StaticContent, ContactMessage, BatchImportation
from django.utils import timezone
from enum import StrEnum
from datetime import date, timedelta
from django.db.models import Q, F

from django.urls import reverse_lazy
from django.utils.translation import gettext_lazy as _
from django.utils.translation import activate, get_language_info
import django_filters
from django.core.paginator import Paginator, EmptyPage, PageNotAnInteger
from django.contrib.auth.decorators import login_required

from django.contrib import messages
from django.contrib.messages.views import SuccessMessageMixin

from .calendar import CalendarMonth, CalendarWeek 
from .extractors import ExtractorAllURLs

from .celery import app as celery_app, import_events_from_json, import_events_from_url

import unicodedata


def get_event_qs(request):
    if request.user.is_authenticated:
        return Event.objects.filter(~Q(status=Event.STATUS.TRASH))
    else:
        return Event.objects.filter(status=Event.STATUS.PUBLISHED)




class CategoryCheckboxSelectMultiple(forms.CheckboxSelectMultiple):
    template_name = 'agenda_culturel/forms/category-checkbox.html'
    option_template_name = 'agenda_culturel/forms/checkbox-option.html'

class TagCheckboxSelectMultiple(forms.CheckboxSelectMultiple):
    template_name = 'agenda_culturel/forms/tag-checkbox.html'
    option_template_name = 'agenda_culturel/forms/checkbox-option.html'


class EventFilter(django_filters.FilterSet):
    tags = django_filters.MultipleChoiceFilter(label="Étiquettes", 
        choices=[(t, t) for t in Event.get_all_tags()], 
        lookup_expr='icontains', 
        field_name="tags", 
        widget=TagCheckboxSelectMultiple)


    category = django_filters.ModelMultipleChoiceFilter(label="Catégories", 
        field_name="category__id", 
        to_field_name='id', 
        queryset=Category.objects.all(), 
        widget=CategoryCheckboxSelectMultiple)


    class Meta:
        model = Event
        fields = ["category", "tags"]

    def get_url(self):
        if isinstance(self.form.data, QueryDict):
            return self.form.data.urlencode() 
        else:
            print(self.form.data)
            return ""

    def get_url_without_filters(self):
        return self.request.build_absolute_uri().split("?")[0]

    def get_categories(self):
        return self.form.cleaned_data["category"]

    def get_tags(self):
        return self.form.cleaned_data["tags"]

    def is_active(self, only_categories=False):
        if only_categories:
            return len(self.form.cleaned_data["category"]) != 0
        else:    
            return len(self.form.cleaned_data["category"]) != 0 or len(self.form.cleaned_data["tags"]) != 0

    def is_selected(self, cat):
        return cat in self.form.cleaned_data["category"]


def mentions_legales(request):

    context = { "title": "Mentions légales", "static_content": "mentions_legales", "url_path": reverse_lazy("mentions_legales") }
    return render(request, 'agenda_culturel/page-single.html', context)


def home(request):
    return week_view(request, home=True)

def month_view(request, year = None, month = None):
    now = date.today()
    if year is None:
        year = now.year
    if month is None:
        month = now.month

    filter = EventFilter(request.GET, queryset=get_event_qs(request), request=request)
    cmonth = CalendarMonth(year, month, filter)
    

    context = {"year": year, "month": cmonth.get_month_name(), "calendar": cmonth, "filter": filter }
    return render(request, 'agenda_culturel/page-month.html', context)


def week_view(request, year=None, week=None, home=False):
    now = date.today()
    if year is None:
        year = now.year
    if week is None:
        week = now.isocalendar()[1]

    filter = EventFilter(request.GET, queryset=get_event_qs(request), request=request)
    cweek = CalendarWeek(year, week, filter)

    context = {"year": year, "week": week, "calendar": cweek, "filter": filter }
    if home:
        context["home"] = 1
    return render(request, 'agenda_culturel/page-week.html', context)


def day_view(request, year = None, month = None, day = None):
    now = date.today()
    if year is None:
        year = now.year
    if month is None:
        month = now.month
    if day is None:
        day = now.day

    day = date(year, month, day)
    
    filter = EventFilter(request.GET, get_event_qs(request), request=request)
    events = filter.qs.filter((Q(start_day__lte=day) & (Q(end_day__gte=day)) | Q(start_day=day))).order_by("start_day", F("start_time").desc(nulls_last=True))

    context = {"day": day, "events": events, "filter": filter}
    return render(request, 'agenda_culturel/page-day.html', context)


def view_tag(request, t):
    events = Event.objects.filter(tags__contains=[t]).order_by("start_day", "start_time")
    context = {"tag": t, "events": events}
    return render(request, 'agenda_culturel/tag.html', context)


def tag_list(request):
    def remove_accents(input_str):
        nfkd_form = unicodedata.normalize('NFKD', input_str)
        return u"".join([c for c in nfkd_form if not unicodedata.combining(c)])

    tags = Event.get_all_tags()    
    context = {"tags": sorted(tags, key=lambda x: remove_accents(x).lower())}
    return render(request, 'agenda_culturel/tags.html', context)


class StaticContentCreateView(LoginRequiredMixin, CreateView):
    model = StaticContent
    fields = ['text']

    def form_valid(self, form):
         form.instance.name = self.request.GET["name"]
         form.instance.url_path = self.request.GET["url_path"]
         return super().form_valid(form)


class StaticContentUpdateView(SuccessMessageMixin, LoginRequiredMixin, UpdateView):
    model = StaticContent
    fields = ['text']
    success_message = _('The static content has been successfully updated.')


class EventUpdateView(SuccessMessageMixin, LoginRequiredMixin, UpdateView):
    model = Event
    form_class = EventForm
    success_message = _('The event has been successfully modified.')

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs['is_authenticated'] = self.request.user.is_authenticated
        return kwargs


class EventDeleteView(SuccessMessageMixin, LoginRequiredMixin, DeleteView):
    model = Event
    success_url = reverse_lazy('moderation')
    success_message = _('The event has been successfully deleted.')


class EventDetailView(UserPassesTestMixin, DetailView):
    model = Event
    template_name = "agenda_culturel/page-event.html"

    def test_func(self):
        return self.request.user.is_authenticated or self.get_object().status == Event.STATUS.PUBLISHED

@login_required(login_url="/accounts/login/")
def change_status_event(request, pk, status):
    event = get_object_or_404(Event, pk=pk)

    if request.method == 'POST':
        event.status = Event.STATUS(status)
        event.save(update_fields=["status"])
        messages.success(request, _("The status has been successfully modified."))

        if request.user.is_authenticated:
            return HttpResponseRedirect(event.get_absolute_url())
        else:
            return HttpResponseRedirect(reverse_lazy("home"))
    else:
        cancel_url = request.META.get('HTTP_REFERER', '')
        if cancel_url == "":
            cancel_url = reverse_lazy("home")
        return render(request, 'agenda_culturel/event_confirm_change_status.html', {"status": status, "event": event, "cancel_url": cancel_url})

def import_from_url(request):

    import logging
    logger = logging.getLogger(__name__)

    if request.method == 'POST' and "title" in request.POST:
        form = EventForm(request.POST)
        if form.is_valid():
            new_event = form.save()
            if request.user.is_authenticated:
                messages.success(request, _("The event is saved."))
                return HttpResponseRedirect(new_event.get_absolute_url())
            else:
                messages.success(request, _("The event has been submitted and will be published as soon as it has been validated by the moderation team."))
                return HttpResponseRedirect(reverse("home"))
        else:
            return render(request, 'agenda_culturel/event_form.html', context={'form': form })

    else:
        form = EventSubmissionForm()
        
        initial = {"start_day": date.today() + timedelta(days=1),
                   "start_time": "20:00", 
                   "end_time": "22:00"}

        form_event = EventForm(initial=initial)

        if request.method == 'POST':
            form = EventSubmissionForm(request.POST)


            if form.is_valid():
                cd = form.cleaned_data
                url = cd.get('url')

                url = ExtractorAllURLs.clean_url(url)

                existing = Event.objects.filter(uuids__contains=[url])

                if len(existing) == 0:
                    event = ExtractorAllURLs.extract(url)
                    if event != None:
                        form = EventForm(instance=event)
                        messages.success(request, _("The event has been successfully extracted, and you can now submit it after modifying it if necessary."))
                        return render(request, 'agenda_culturel/event_form.html', context={'form': form })
                    else:
                        form = EventForm(initial={'reference_urls': [url]})
                        messages.error(request, _("Unable to extract an event from the proposed URL. Please use the form below to submit the event."))
                        return render(request, 'agenda_culturel/import.html', context={'form': form, 'form_event': form_event})
                else:
                    published = [e for e in existing if e.status == Event.STATUS.PUBLISHED]
                    drafts = [e for e in existing if e.status == Event.STATUS.DRAFT]
                    trash = [e for e in existing if e.status == Event.STATUS.TRASH]

                    if request.user.is_authenticated or len(published) > 1:
                        event = published[0] if len(published) > 1 else existing[0]
                        messages.info(request, _("This URL has already been submitted, and you can find the event below."))
                        return HttpResponseRedirect(event.get_absolute_url())
                    else:
                        if len(drafts) > 0:
                            messages.info(request, _("This URL has already been submitted, but has not been selected for publication by the moderation team."))
                        elif len(trash) > 0:
                            messages.info(request, _("This URL has already been submitted and is awaiting moderation."))
        

        return render(request, 'agenda_culturel/import.html', context={'form': form, 'form_event': form_event})


class EventFilterAdmin(django_filters.FilterSet):
    status = django_filters.MultipleChoiceFilter(choices=Event.STATUS.choices, widget=forms.CheckboxSelectMultiple)

    class Meta:
        model = Event
        fields = ['status']




class ContactMessageCreateView(SuccessMessageMixin, CreateView):
    model = ContactMessage
    template_name = "agenda_culturel/contactmessage_create_form.html"
    fields = ['subject', 'name', 'email', 'message']

    success_url = reverse_lazy('home')
    success_message = _('Your message has been sent successfully.')


class ContactMessageUpdateView(SuccessMessageMixin, LoginRequiredMixin, UpdateView):
    model = ContactMessage
    template_name = "agenda_culturel/contactmessage_moderation_form.html"
    fields = ('closed', 'comments')
    
    success_message = _('The contact message properties has been successfully modified.')

    success_url = reverse_lazy('contactmessages')


    def get_form_kwargs(self):
        """Return the keyword arguments for instantiating the form."""
        kwargs = super().get_form_kwargs()
        if hasattr(self, 'object'):
            kwargs.update({'instance': self.object})
        return kwargs



class ContactMessagesFilterAdmin(django_filters.FilterSet):
    closed = django_filters.MultipleChoiceFilter(label="Status", choices=((True, _("Closed")), (False, _("Open"))), widget=forms.CheckboxSelectMultiple)
    class Meta:
        model = ContactMessage
        fields = ['closed']



@login_required(login_url="/accounts/login/")
def moderation(request):
    filter = EventFilterAdmin(request.GET, queryset=Event.objects.all().order_by("-created_date"))
    paginator = Paginator(filter.qs, 10)
    page = request.GET.get('page')

    try:
        response = paginator.page(page)
    except PageNotAnInteger:
        response = paginator.page(1)
    except EmptyPage:
        response = paginator.page(paginator.num_pages)

    return render(request, 'agenda_culturel/moderation.html', {'filter': filter, 'paginator_filter': response} )

@login_required(login_url="/accounts/login/")
def contactmessages(request):
    filter = ContactMessagesFilterAdmin(request.GET, queryset=ContactMessage.objects.all().order_by("-date"))
    paginator = Paginator(filter.qs, 10)
    page = request.GET.get('page')

    try:
        response = paginator.page(page)
    except PageNotAnInteger:
        response = paginator.page(1)
    except EmptyPage:
        response = paginator.page(paginator.num_pages)

    return render(request, 'agenda_culturel/contactmessages.html', {'filter': filter, 'paginator_filter': response} )



class SimpleSearchEventFilter(django_filters.FilterSet):
    q = django_filters.CharFilter(method='custom_filter', label=_("Search"))

    def custom_filter(self, queryset, name, value):
        search_query = SearchQuery(value, config='french')
        qs = queryset.filter(
            Q(title__icontains=value) | Q(location__icontains=value) | Q(description__icontains=value))
        for f in ["title", "location", "description"]:
            params = { f + "_hl": SearchHeadline(f,
                search_query,
                start_sel="<span class=\"highlight\">",
                stop_sel="</span>",
                config='french')}
            qs = qs.annotate(**params)
        return qs

    class Meta:
        model = Event
        fields = ['q']


class SearchEventFilter(django_filters.FilterSet):
    tags = django_filters.CharFilter(lookup_expr='icontains')
    title = django_filters.CharFilter(method="hl_filter_contains")
    location = django_filters.CharFilter(method="hl_filter_contains")
    description = django_filters.CharFilter(method="hl_filter_contains")
    start_day = django_filters.DateFromToRangeFilter(widget=django_filters.widgets.RangeWidget(attrs={'type': 'date'}))

    o = django_filters.OrderingFilter(
        # tuple-mapping retains order
        fields=(
            ('title', 'title'),
            ('description', 'description'),
            ('start_day', 'start_day'),
        ),
    )

    def hl_filter_contains(self, queryset, name, value):

        # first check if it contains
        filter_contains = { name + "__contains": value }
        queryset = queryset.filter(**filter_contains)

        # then hightlight the result
        search_query = SearchQuery(value, config='french')
        params = { name + "_hl": SearchHeadline(name,
                search_query,
                start_sel="<span class=\"highlight\">",
                stop_sel="</span>",
                config='french')}
        return queryset.annotate(**params)


    class Meta:
        model = Event
        fields = ['title', 'location', 'description', 'category', 'tags', 'start_day']



def event_search(request, full=False):

    if full:
        filter = SearchEventFilter(request.GET, queryset=get_event_qs(request).order_by("-start_day"))
    else:
        filter = SimpleSearchEventFilter(request.GET, queryset=get_event_qs(request).order_by("-start_day"))


    paginator = Paginator(filter.qs, 10)
    page = request.GET.get('page')

    try:
        response = paginator.page(page)
    except PageNotAnInteger:
        response = paginator.page(1)
    except EmptyPage:
        response = paginator.page(paginator.num_pages)

    return render(request, 'agenda_culturel/search.html', {'filter': filter, 
        'has_results': len(request.GET) != 0 or (len(request.GET) > 1 and "page" in request.GET),
        'paginator_filter': response, 
        'full': full})

def event_search_full(request):
    return event_search(request, True)


#########################
## batch importations
#########################

@login_required(login_url="/accounts/login/")
def imports(request):
    paginator = Paginator(BatchImportation.objects.all().order_by("-created_date"), 10)
    page = request.GET.get('page')

    try:
        response = paginator.page(page)
    except PageNotAnInteger:
        response = paginator.page(1)
    except EmptyPage:
        response = paginator.page(paginator.num_pages)

    return render(request, 'agenda_culturel/imports.html', {'paginator_filter': response} )


class BatchImportationCreateView(SuccessMessageMixin, LoginRequiredMixin, CreateView):
    model = BatchImportation

    success_url = reverse_lazy('imports')
    success_message = _('The import has been run successfully.')

    form_class = BatchImportationForm


    def form_valid(self, form):

        # run import
        if "json" in form.data and form.data["json"] is not None and form.data["json"].strip() != "":
            result = import_events_from_json.delay(form.data["json"])
        else:
            result = import_events_from_url.delay(form.data["source"], form.data["browsable_url"])

        # update the object with celery_id
        form.instance.celery_id = result.id

        return super().form_valid(form)



@login_required(login_url="/accounts/login/")
def cancel_import(request, pk):
    import_process = get_object_or_404(BatchImportation, pk=pk)

    if request.method == 'POST':
        celery_app.control.revoke(import_process.celery_id)

        import_process.status = BatchImportation.STATUS.CANCELED
        import_process.save(update_fields=["status"])

        messages.success(request, _("The import has been canceled."))
        return HttpResponseRedirect(reverse_lazy("imports"))
    else:
        cancel_url = reverse_lazy("imports")
        return render(request, 'agenda_culturel/cancel_import_confirm.html', {"object": import_process, "cancel_url": cancel_url})

